'''Utils for managing wpa supplicant files.
'''
import os
import glob
import logging
from shutil import copyfile
from . import util


logger = logging.getLogger(__name__)

def set_ap_path(path):
    Wpa.ap_path = path

def connect(ssid, verify=False):
    return Wpa(ssid).connect() and (not verify or verify_ssid(ssid))

def verify_ssid(ssid):
    return Wpa().ssid == ssid


class Wpa:
    ap_path = '/etc/wpa_supplicant/aps'
    WPA_PATH = "/etc/wpa_supplicant/wpa_supplicant.conf"
    def __init__(self, ssid=None, path=None, iface='wlan0', ap_path=None):
        self.ap_path = ap_path or self.ap_path
        self.iface = iface
        self.path = path or (ssid_path(ssid, self.ap_path) if ssid else self.WPA_PATH)
        self.ssid = ssid or self.info.get('ssid')

    def connect(self, backup=True, restart=True):
        '''Set ap as current wpa_supplicant.'''
        wpa = Wpa()
        if wpa.ssid != self.ssid:
            if backup:
                wpa.backup()
            return (
                self.copy(self.WPA_PATH)
                and not restart or util.restart_iface(self.iface))
        return True

    @property
    def exists(self):
        return self.path and os.path.exists(self.path)

    def copy(self, dest):
        '''Copy wpa_supplicant to destination.'''
        if self.path != dest and self.exists:
            copyfile(self.path, dest)
        return True

    def create(self, **kw):
        '''Generate wpa supplicant.'''
        generate_wpa_config(self.ssid, **kw)

    def backup(self, force=True):
        '''Make sure that the current wifi network is present in aps/.
        '''
        if force or self.ssid not in ssids_from_dir(self.ap_path):
            self.copy(ssid_path(self.ssid, self.ap_path))
            return True
        return False

    @property
    def info(self):
        '''Parse the wpa_supplicant.conf file and return key value pairs.'''
        DROP_KEYS = ('network',)
        if self._info is None:
            if not self.exists:
                return {}
            with open(self.path) as wpa_conf:
                self._info = {
                    x[0].strip(): x[1].strip('"\' ') for x in
                    (l.split("=", 1) for l in wpa_conf.read().splitlines())
                    if len(x) > 1 and x[0] not in DROP_KEYS
                }
                self._info['password'] = self._info.get('psk')
        return self._info
    _info = None

    def __getattr__(self, k):
        try:
            return self.info[k]
        except KeyError as e:
            raise AttributeError(e)

    def _summary(self):
        import json
        cfg = util.mask_dict_values(
                self.info, 'password', drop=('psk', 'ctrl_interface', 'update_config'))
        return json.dumps(cfg, indent=4 if len(cfg) > 3 else None, sort_keys=True)

    def summary(self):
        print(self._summary())



def ssid_path(ssid, ap_path=None):
    return os.path.join(ap_path or Wpa.ap_path, f'{ssid}.conf')


def ssids_from_dir(ap_path=None, pat='*.conf'):
    '''Get file name -> file path mapping for files in a directory.
    e.g.: `{file: ap_path/file.ext for f in glob(ap_path)}`'''
    return {
        os.path.splitext(os.path.basename(f))[0]: f
        for f in glob.glob(os.path.join(ap_path or Wpa.ap_path, pat))}


def sync_aps(repo_path, ap_path=None, force=True, backup=True):
    '''Copy aps from network/aps to the trusted aps path.'''
    ap_path = ap_path or Wpa.ap_path
    os.makedirs(ap_path, exist_ok=True)
    fnames_repo = ssids_from_dir(repo_path)
    fnames_aps = ssids_from_dir(ap_path)
    for fn in (fnames_repo if force else set(fnames_repo) - set(fnames_aps)):
        copyfile(fnames_repo[fn], os.path.join(ap_path, fn + '.conf'))
    if backup:
        Wpa().backup()



'''

Create Wpa Supplicant File

'''

def generate_wpa_config(
        ssid, password=None, kind='basic', group='netdev',
        country='US', askpass=False, ap_path=None, **kw):
    '''Create the wpa_supplicant configuration file
    in the `./aps/{ssid}.conf`

    Arguments:
        ssid (str): Access point to add credentials for.
        password (str, optional): The

    Returns:
        status (bool): True if no exception was raised when creating
    '''
    logger.debug("Creating config for: " + str(ssid))
    password = password or kw.pop('psk', None)
    if askpass and not password:
        import getpass
        password = getpass.getpass()

    if kind == 'basic':
        network = (
            dict(ssid=ssid, psk=password) if password else
            dict(ssid, key_mgmt='NONE'))
    elif kind == 'edu':
        network = dict(
            ssid=ssid, proto='RSN', key_mgmt='WPA-EAP',
            pairwise='CCMP', phase2="auth=MSCHAPV2",
            auth_alg='OPEN', eap='PEAP',
            identity=kw.pop('identity'),
            password=password)
    else:
        raise ValueError('Unknown wpa config format "{}"'.format(kind))

    tmpl = '''
ctrl_interface=DIR=/var/run/wpa_supplicant GROUP={group}
update_config=1
country={country}
network={{
{network}
}}'''.strip()

    with open(ensure_dir(ssid_path(ssid, ap_path=ap_path)), 'w') as f:
        f.write(tmpl.format(
            group=group, country=country,
            network=util.indent(_wpa_keys(**network, **kw), 2)))
    return True


def ensure_dir(path):
    dname = os.path.dirname(path)
    if dname:
        os.makedirs(dname, exist_ok=True)
    return path


def _wpa_keys(*a, sep='\n', **kw):
    return sep.join(list(map(str, a)) + [
        '{}={}'.format(k, '{!r}'.format(v))
        for k, v in kw.items()
    ])



import os
import re
import sys
import glob
import time
# import shlex
import functools
import fnmatch
import subprocess
import ifcfg
import yaml
from . import iw, wpasup, util
# import cachetools.func
from .util import internet_connected

import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)  # INFO



def _debug_args(func):
    def inner(*a, **kw):
        print('calling', func.__name__, a, kw)
        return func(*a, **kw)
    return inner


def _maybe_load_yaml(func):  # this is so that we can pull parameters from a yaml file
    #@_debug_args
    @functools.wraps(func)
    def inner(self, config=None, **kw):
        if isinstance(config, str):
            kw = dict(yaml.load(config), **kw)
            config = kw.pop('config', None)
        return func(self, config, **kw)
    return inner


class NetSwitch:
    '''

    Example:

    ```
    NetSwitch([
        {'interface': 'wlan*', 'ssids': 's0nycL1f3l1ne'},
        {'interface': 'eth*'},
        {'interface': 'ppp*'},
        {'interface': 'wlan*'},
    ])
    ```

    assume:
     - ifaces: wlan0, wlan1, ppp0
     - trusted ssids: nyu, nyu-legacy

    Procedure:
     - check wlan1 for s0nycL1f3l1ne
     - check wlan0 for s0nycL1f3l1ne
     - check for eth0 and internet connected thru interface
     - check for ppp0 and internet connected thru interface
        - should we control wvdial or whatevs in this ???
     - check wlan1 for [nyu, nyu-legacy]
     - check wlan0 for [nyu, nyu-legacy]
     - check if internet is already connected thru any interface

    '''
    wlans = {}  # cache at a class level - move if iw.Wlan gets more specific
    @_maybe_load_yaml
    def __init__(self, config=None, lifeline=os.getenv('LIFELINE_SSID'), networks=None, ap_path=None, restart_missing_ip=False):
        config = config or ['eth*', 'ppp*', 'wlan*']
        self.config = [
            {'interface': c} if isinstance(c, str) else c
            for c in (config if isinstance(config, (list, tuple)) else [config])
        ]
        self.restart_missing_ip = restart_missing_ip
        if lifeline:
            logger.info('Using lifeline network: {}'.format(lifeline))
            self.config = [{'interface': 'wlan*', 'ssids': lifeline}] + self.config
        if ap_path:
            wpasup.set_ap_path(ap_path)
        for w in networks or ():
            wpasup.generate_wpa_config(**w)

    def __str__(self):
        return self._summary()

    def check(self, test=False):
        '''Check internet connections and interfaces. Return True if connected.'''
        interfaces = ifcfg.interfaces()
        logger.info('Interfaces: {}'.format(list(interfaces)))
        for cfg in self.config:
            # check if any matching interfaces are available
            ifaces = [
                i for i in interfaces
                if fnmatch.fnmatch(i, cfg['interface'])]

            logger.info('Iface {} - matches {}'.format(cfg['interface'], ifaces))
            # try to connect in order of wlan1, wlan0
            restart_missing = cfg.get('restart_missing_ip', self.restart_missing_ip)
            for iface in sorted(ifaces, reverse=True):
                if restart_missing and not interfaces[iface].get('inet'):
                    util.ifup(iface)
                    #logger.info('ifup {} {}'.format(iface, interfaces[iface].get('inet')))
                if self.connect(iface, cfg) and internet_connected(iface):
                    return True
        # check if internet is connected anyways
        return internet_connected()

    def run(self, interval=10):
        self.summary()
        self.check()
        while True:
            time.sleep(interval)
            check = self.check()
            # logger.debug('Finished check. Connected to internet? {}'.format(check))
        self.summary()

    def connect(self, iface, cfg, **kw):
        '''Connect to an interface.'''
        if fnmatch.fnmatch(iface, 'wlan*'):
            return self.connect_wlan(iface, cfg, **kw)
        return True

    def connect_wlan(self, iface, cfg, test=False):
        # get the wlan connection object
        if iface not in self.wlans:  # cache it for future use
            self.wlans[iface] = iw.WLan(iface=iface)
        wlan = self.wlans[iface]

        # get matching ssids
        ssids = cfg.get('ssids', '*')
        if not isinstance(ssids, (list, tuple)):  # coerce to list of globs
            ssids = [ssids]
        ssids = [
            os.path.splitext(os.path.basename(s))[0] for pat in ssids
            for s in glob.glob(wpasup.ssid_path(pat))]

        # check for available ssids and take best one
        ssid = wlan.select_best_ssid(ssids)
        if not ssid:
            logger.info('No wifi matches.')
            return

        connected = test or wpasup.connect(ssid, verify=True)
        logger.info(
            'AP ({}) Connected? {}. [{}]'.format(
                ssid, connected, iface))
        return connected

    def __getitem__(self, index):
        return self.__class__(self.config[index])

    def _summary(self):
        import json
        return '\n' + '\n'.join((
            '-'*50,
            'Current Network:',
            wpasup.Wpa()._summary(),
            '', 'Trusted APs: {}'.format(', '.join(wpasup.ssids_from_dir(wpasup.Wpa.ap_path))),
            '', 'Config: {}'.format(self.config),
            # '', 'Available Networks:',
            # json.dumps(ifcfg.interfaces(), indent=4, sort_keys=True),
            '', 'Interfaces:',
            '\n'.join('\t{:<16}: {:>16} {:>18}'.format(
                str(d.get('device')), str(d.get('inet')), str(d.get('ether')),
            ) for d in ifcfg.interfaces().values()),
            # json.dumps(ifcfg.interfaces(), indent=4, sort_keys=True),
            '-'*50,
        )) + '\n'

    def summary(self):
        logger.info(self._summary())


#@_debug_args
# @functools.wraps(NetSwitch)
def run(config=None, interval=20, **kw):
    witch = NetSwitch(config, **kw)
    try:
        witch.run(interval=interval)
    except KeyboardInterrupt:
        print('Interrupted.')
    return witch


if __name__ == '__main__':
    NetSwitch([
        {'interface': 'wlan*', 'ssids': ['s0nycL1f3l1ne']},
        {'interface': 'eth*'},
        {'interface': 'ppp*'},
        {'interface': 'wlan*', 'ssids': '*'},
    ])



import time
from collections import Counter
from access_points import get_scanner
import logging


logger = logging.getLogger(__name__)


class WLan:
    def __init__(self, iface='wlan0'):
        self.iface = iface
        self.wifi_scanner = get_scanner(iface)

    def scan(self, trusted=None):
        aps = self.wifi_scanner.get_access_points()
        aps = sorted(aps, key=lambda ap: ap.quality, reverse=True)
        #logger.info('all aps: {}'.format([a.ssid for a in aps]))
        return [ap for ap in aps if ap.ssid in trusted] if trusted else aps

    def ap_available(self, ap):
        '''Check if an ap is available.'''
        return any(1 for ap_i in self.scan() if ap in ap_i.ssid)

    def select_best_ssid(self, ssids=None, nmin=3, return_all=False, n_single=4, **kw):
        # handle special cases
        if len(ssids) == 1:
            logger.debug('Checking for network: {}'.format(ssids[0]))
            return any(self.ap_available(ssids[0]) for _ in range(n_single)) and ssids[0]
        # select best
        top_seen, all_seen = self._get_top_ssids(ssids, **kw)
        most_common = Counter(top_seen).most_common(1)
        ap, count = most_common[0] if most_common else (None, -1)
        out_ap = count >= nmin and ap
        if ap and not out_ap:
            logger.debug('AP ({}) was available but not strong enough ({}/{}).'.format(ap, count, nmin))
        return (out_ap, all_seen) if return_all else out_ap

    def _get_top_ssids(self, ssids=None, nscans=5, throttle=1, timeout=30):
        all_seen, top_seen = set(), []
        t0 = time.time()
        #logger.debug('Selecting best network from: {}'.format(ssids or 'all'))
        while len(top_seen) < nscans:
            sids = [ap.ssid for ap in self.scan()]
            trusted = [s for s in sids if s in ssids] if ssids else sids
            logger.info('Scan {} - trusted: {}, all={}'.format(
                len(top_seen), trusted, len(sids)))
            all_seen.update(trusted)
            top_seen.extend(trusted[:1])
            # throttle and timeout
            time.sleep(throttle)
            if timeout and time.time() - t0 >= timeout:
                break
        return top_seen, all_seen



import ifcfg
from . import util
from .core import *
from .iw import *
from .wpasup import *


def get_ifaces(*ifaces):
    avail = ifcfg.interfaces()
    matches = util.matches(ifaces, avail)
    return {iface: avail[iface] for iface in matches}


def get_aps(*ifaces):
    '''List available APs for an interface.'''
    return {
        iface: WLan(iface).scan()
        for iface in get_ifaces(*(ifaces or ('wlan*',)))
    }


def get_ip(*ifaces, key='inet'):
    '''Get the ip of an interface.'''
    return get_iface_info(key, *ifaces)


def get_iface_info(key, *ifaces):
    '''Get an attribute for each matching interface.'''
    info = {iface: d.get(key) for iface, d in get_ifaces(*ifaces).items()}
    return {k: v for k, v in info.items() if v is not None}


def cli():
    import fire
    fire.Fire({
        'ip': get_ip,
        'aps': get_aps,
        'iface': get_ifaces,
        'connected': internet_connected,
        'restart': util.restart_iface,
        'wpa': Wpa,
        'run': run,
    })



'''

source-directory /etc/network/interfaces.d

auto lo
iface lo inet loopback

auto eth0
iface eth0 inet manual

allow-hotplug ppp0
iface ppp0 inet wvdial
post-up echo "cellular (ppp0) is online"

allow-hotplug wlan0
iface wlan0 inet manual
wpa-roam /etc/wpa_supplicant/wpa_supplicant.conf

allow-hotplug wlan1
iface wlan1 inet manual
wpa-roam /etc/wpa_supplicant/wpa_supplicant.conf
'''
import re
import functools

def _expand_wildcard(*rng):
    def outer(func):
        @functools.wraps(func)
        def inner(i='*', *a, **kw):
            return (
                '\n'.join(func(i, *a, **kw) for i in range(*rng))
                if i == '*' else func(i, *a, **kw))
        return inner
    return outer


# interfaces

def source_dir(fname='/etc/network/interfaces.d'):
    return 'source-directory {}'.format(fname)

def lo(**kw):
    return _iface('lo', method='loopback', **kw)

@_expand_wildcard(2)
def eth(i=0, **kw):
    return _iface('eth{}'.format(i), **kw)

@_expand_wildcard(2)
def wlan(i=0, wpa='/etc/wpa_supplicant/wpa_supplicant.conf', roam=True, **kw):
    kw['wpa_{}'.format('roam' if roam else 'conf')] = wpa
    return _iface('wlan{}'.format(i), hotplug=True, **kw)

@_expand_wildcard(2)
def ppp(i=0, method='wvdial', **kw):
    name = 'ppp{}'.format(i)
    return _iface(
        name, hotplug=True, method=method,
        post_up='echo "cellular ({}) is online"'.format(name), **kw)


IFACES = {'wlan': wlan, 'eth': eth, 'ppp': ppp}

def iface(name, *a, **kw):
    if callable(name):
        return name(*a, **kw)
    matches = re.search(r'([A-z]+)([\d\*]*)', name)
    name, i = matches.groups()
    return IFACES[name.lower()](i or '*', *a, **kw)

# utils

def _iface(name, method='manual', hotplug=False, static=None, **kw):
    return '''
{allow} {name}
iface {name} inet {method}
{extra}
'''.format(
    name=name, allow='allow-hotplug' if hotplug else 'auto', method=method,
    extra='\n'.join(l for k, v in kw.items() for l in _cfg_lines(k, v)))

def _cfg_lines(name, value):
    name = name.replace('_', '-')
    for v in value if isinstance(value, (list, tuple)) else [value]:
        yield '{} {}'.format(name, v)


# config formats

def build_file(*ifaces):
    return '\n'.join([source_dir(), lo()] + [
        ifc() if callable(ifc) else ifc for ifc in ifaces
    ])

def default():
    return build_file(eth, ppp, wlan)

def from_config(config=None):
    config = [
        {'interface': 'wlan*', 'ssids': 's0nycL1f3l1ne'},
        {'interface': 'eth*'},
        {'interface': 'ppp*'},
        {'interface': 'wlan*'},
    ]
    return build_file(*(
        iface(c['interface'], **c.get('etc', {}))
        for c in config
    ))


if __name__ == '__main__':
    import fire
    fire.Fire()



import sys
import fnmatch
import subprocess
import logging

logger = logging.getLogger(__name__)


def matches(pattern, items):
    patterns = list(flatten(pattern)) or ['*']
    return [
        item for item in items
        if any(
            fnmatch.fnmatch(item, pattern)
            for pattern in patterns
        )
    ]


def flatten(*items):
    '''Yield items from any nested iterable.'''
    for item in items:
        if isinstance(item, (list, tuple)):
            yield from flatten(*item)
        else:
            yield item



def mask_dict_values(dct, *keys, ch='*', drop=None):
    return {
        k: '*'*len(v) if k in keys else v
        for k, v in dct.items()
        if not drop or k not in drop
    }


def indent(txt, n=1, w=2):
    return ''.join(' '*w*n + l for l in txt.splitlines(keepends=True))


# internet

def internet_connected(iface=None, n=3):
    '''Check if we're connected to the internet (optionally, check a specific interface `iface`)'''
    try:
        result = subprocess.run(
            "ping {} -c {} 8.8.8.8".format(
                '-I {}'.format(iface) if iface else '', n),
            capture_output=True, check=True, shell=True)
        #logger.info('connected? {} - {}, {}'.format(not result.stderr, result.stdout, result.stderr))
        return not result.stderr
    except subprocess.CalledProcessError as e:
        #logger.debug(e.stderr.decode('utf-8'))
        pass


# ifup / ifdown

def _ifupdown_(cmd, name, sleep=1, force=True):
    try:
        subprocess.run(
            #'if{} {} {} && sleep {}'.format(cmd, name, force*'--force', sleep),
            'ifconfig {} {} && sleep {}'.format(name, cmd, sleep),
            check=True, capture_output=True, shell=True)
    except subprocess.CalledProcessError as e:
        logger.error(e.stderr and e.stderr.decode())
        return False
    return True

def ifup(name, sleep=1, force=True):
    return _ifupdown_('up', name, sleep=sleep, force=force)

def ifdown(name, sleep=1, force=True):
    return _ifupdown_('down', name, sleep=sleep, force=force)

def restart_iface(name=None, sleep=3):
    '''Restart the specified network interface. Returns True if restarted without error.'''
    logger.info("Restarting Interface: {}".format(name))
    went_down = ifdown(name, sleep)
    back_up = ifup(name, sleep)
    return went_down and back_up



import glob
import itertools
import serial
from contextlib import contextmanager


DEFAULT_DEVICE_PATTERN = '/dev/ttyUSB*'
DEFAULT_DEVICE = '/dev/ttyUSB2'
DEFAULT_BAUDRATE = 115200
DEFAULT_TIMEOUT = 0.01
OK = 'OK'



class Cell(serial.Serial):
    def __init__(self, device=DEFAULT_DEVICE, baudrate=DEFAULT_BAUDRATE,
                 timeout=DEFAULT_TIMEOUT, **kw):
        super(Cell, self).__init__(device, baudrate, timeout=timeout, **kw)

    def send(self, cmd, raw=False, end='\r', length=64):
        self.write((cmd + end).encode())
        msg = self.read(length).decode("utf-8")
        lines = msg.splitlines()
        if lines and OK in lines:
            if not raw:
                msg = '\n'.join(
                    l for l in lines if l not in ('', cmd, OK)) or True
            return msg

    def sendn(self, *cmds, **kw):
        return list(itertools.takewhile(
            lambda x: x is not None,
            (self.send(c, **kw) for c in cmds)))

    def ping(self):
        return self.send('AT')

    @property
    def quality(self):
        '''Return signal quality (in dB).'''
        msg = self.ping() and self.send("AT+CSQ")
        if msg:
            value = int(msg.split(': ')[1].split(',')[0].strip())
            return 2 * value - 112

    def chat(self):
        try:
            print('Opened Chat:')
            while True:
                cmd = input('*** ')
                if cmd.lower() == 'exit':
                    break
                print(self.send(cmd))
        except KeyboardInterrupt:
            pass



def devices(device_pattern=DEFAULT_DEVICE_PATTERN):
    return glob.glob(device_pattern)

@contextmanager
def cell_device(device_pattern=DEFAULT_DEVICE_PATTERN, *a, **kw):
    for dev in glob.glob(device_pattern):
        with Cell(dev, *a, **kw) as cell:
            if cell.send('ATZ'):
                yield cell
                return
    raise OSError('No device found matching {}.'.format(device_pattern))

def signal_strength(com_device=DEFAULT_DEVICE):
    with cell_device(com_device) as cell:
        return cell.quality


if __name__ == '__main__':
    import fire
    fire.Fire(cell_device)
